try:
    import bankroll.brokers.ibkr as ibkr
except ImportError:
    ibkr = None  # type: ignore

try:
    import bankroll.brokers.schwab as schwab
except ImportError:
    schwab = None  # type: ignore

try:
    import bankroll.brokers.fidelity as fidelity
except ImportError:
    fidelity = None  # type: ignore

try:
    import bankroll.brokers.vanguard as vanguard
except ImportError:
    vanguard = None  # type: ignore



import logging
from argparse import ArgumentParser, Namespace
from itertools import chain
from typing import Callable, Dict, Iterable, List, Optional

from progress.bar import Bar  # type: ignore

import bankroll.analysis as analysis
from bankroll.broker import AccountAggregator
from bankroll.broker.configuration import (
    Configuration,
    Settings,
    addSettingsToArgumentGroup,
)
from bankroll.marketdata import MarketConnectedAccountData, MarketDataProvider
from bankroll.model import Activity, Cash, Instrument, Position, Stock, Trade

from .brokers import *
from .configuration import loadConfig, marketDataProvider

parser = ArgumentParser(
    prog="bankroll",
    add_help=False,
    description="Ingests portfolio and other data from multiple brokerages, and analyzes it.",
    epilog="For more information, or to report issues, please visit: https://github.com/bankroll-py/bankroll",
)

# Add our own help option for consistent formatting.
parser.add_argument(
    "-h", "--help", help="Show this help message and exit.", action="help"
)

parser.add_argument(
    "--lenient",
    help="Attempt to ignore invalid data instead of erroring out. May not be supported for all data sources.",
    default=False,
    action="store_true",
)
parser.add_argument(
    "--no-lenient", dest="lenient", help="Opposite of --lenient.", action="store_false"
)
parser.add_argument(
    "-v",
    "--verbose",
    help="Turns on more logging, for debugging purposes.",
    dest="verbose",
    default=False,
    action="store_true",
)
parser.add_argument(
    "--config",
    help="Path to an INI file specifying configuration options, taking precedence over the default search paths. Can be specified multiple times, with the latest file's settings taking precedence over those previous.",
    action="append",
)

if ibkr:
    ibGroup = parser.add_argument_group(
        "IB", "Options for importing data from Interactive Brokers."
    )
    readIBSettings = addSettingsToArgumentGroup(ibkr.Settings, ibGroup)

if fidelity:
    fidelityGroup = parser.add_argument_group(
        "Fidelity",
        "Options for importing data from local files in Fidelity's CSV export format.",
    )
    readFidelitySettings = addSettingsToArgumentGroup(fidelity.Settings, fidelityGroup)

if schwab:
    schwabGroup = parser.add_argument_group(
        "Schwab",
        "Options for importing data from local files in Charles Schwab's CSV export format.",
    )
    readSchwabSettings = addSettingsToArgumentGroup(schwab.Settings, schwabGroup)

if vanguard:
    vanguardGroup = parser.add_argument_group(
        "Vanguard",
        "Options for importing data from local files in Vanguard's CSV export format.",
    )
    readVanguardSettings = addSettingsToArgumentGroup(vanguard.Settings, vanguardGroup)


def printPositions(accounts: AccountAggregator, args: Namespace) -> None:
    values: Dict[Position, Cash] = {}
    if args.live_value:
        dataProvider = marketDataProvider(accounts)
        if dataProvider:
            values = analysis.liveValuesForPositions(
                accounts.positions(),
                dataProvider=dataProvider,
                progressBar=Bar("Loading market data for positions"),
            )
        else:
            logging.error("Live data connection required to fetch market values")

    for p in sorted(accounts.positions(), key=lambda p: p.instrument):
        print(p)

        if p in values:
            print(f"\tMarket value: {values[p]}")
        elif args.live_value:
            logging.warning(f"Could not fetch market value for {p.instrument}")

        print(f"\tCost basis: {p.costBasis}")

        if args.realized_basis and isinstance(p.instrument, Stock):
            realizedBasis = analysis.realizedBasisForSymbol(
                p.instrument.symbol, activity=accounts.activity()
            )
            print(f"\tRealized basis: {realizedBasis}")


def printActivity(accounts: AccountAggregator, args: Namespace) -> None:
    for t in sorted(accounts.activity(), key=lambda t: t.date, reverse=True):
        print(t)


def printBalances(accounts: AccountAggregator, args: Namespace) -> None:
    print(accounts.balance())


def symbolTimeline(accounts: AccountAggregator, args: Namespace) -> None:
    for entry in reversed(
        list(analysis.timelineForSymbol(args.symbol, accounts.activity()))
    ):
        print(entry)


commands: Dict[str, Callable[[AccountAggregator, Namespace], None]] = {
    "positions": printPositions,
    "activity": printActivity,
    "balances": printBalances,
    "timeline": symbolTimeline,
}

subparsers = parser.add_subparsers(dest="command", help="What to inspect")

positionsParser = subparsers.add_parser(
    "positions", help="Operations upon the imported list of portfolio positions"
)
positionsParser.add_argument(
    "--realized-basis",
    help="Calculate realized basis for stock positions",
    default=False,
    action="store_true",
)
positionsParser.add_argument(
    "--live-value",
    help="Fetch live, mark-to-market value of positions",
    default=False,
    action="store_true",
)

activityParser = subparsers.add_parser(
    "activity", help="Operations upon imported portfolio activity"
)

balancesParser = subparsers.add_parser(
    "balances", help="Operations upon imported portfolio cash balances"
)

timelineParser = subparsers.add_parser(
    "timeline", help="Traces a position and P/L for a symbol over time"
)
timelineParser.add_argument(
    "symbol",
    help="The symbol to look up (multi-part symbols like BRK.B will be normalized so they can be tracked across brokers)",
)


def main() -> None:
    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.INFO)

    config = loadConfig(
        chain(Configuration.defaultSearchPaths, args.config if args.config else [])
    )

    if not args.command:
        parser.print_usage()
        quit(1)

    mergedSettings: Dict[Settings, str] = dict(
        chain(
            readFidelitySettings(config, args).items() if fidelity else [],
            readSchwabSettings(config, args).items() if schwab else [],
            readVanguardSettings(config, args).items() if vanguard else [],
            readIBSettings(config, args).items() if ibkr else [],
        )
    )

    accounts = AccountAggregator.fromSettings(mergedSettings, lenient=args.lenient)
    commands[args.command](accounts, args)


if __name__ == "__main__":
    main()



import math
import pandas as pd  # type: ignore
import numpy as np  # type: ignore
import pyfolio as pf  # type: ignore
from typing import Optional, List, Dict, Iterable, Tuple
from decimal import Decimal
from bankroll.model import *
from bankroll.marketdata import *
from dataclasses import asdict


def etf(portfolio: pd.DataFrame, timezone: str) -> pd.Series:
    """
    Returns a time series representing the investment of $1 in a basket of instruments weighted proportionally by the given weights.

    @param portfolio: A DataFrame of instruments containing open, close, and weight data indexed by Date.
    """
    index = portfolio.index.levels[1].tz_localize(timezone)
    # Initialize a zero'd out T-sized array where T is the length of the date range.
    etf = np.zeros(portfolio.loc["open"].shape[0])
    etf[0] = Decimal(1)  # Initial AUM for this instrument is $1.

    # Initialize an I x T array of holdings over time so we can keep track of how we've divided our AUM
    # across the given instruments over time.
    holds = np.zeros(portfolio.loc["open"].shape)

    for t in range(1, etf.shape[0]):
        portfolio_sum: Decimal = Decimal(0)

        for i in portfolio.columns:
            # Calculate the absolute change of asset i from time t to t - 1.
            change = delta(portfolio, i, t)

            # TODO: Figure out where to get dividend information if it's not calculated into the price.
            dividend = Decimal(0)
            exchange_rate = Decimal(1)

            # Calculate the holdings of instrument i at t - 1 and store it in our holds matrix.
            holdings_t_minus_1 = holdings(
                portfolio, holds, i, t - 1, Decimal(etf[t - 1])
            )
            holds[t - 1][portfolio.columns.get_loc(i)] = holdings_t_minus_1

            portfolio_sum += holdings_t_minus_1 * exchange_rate * (change + dividend)

        # The AUM of the etf at time t is the AUM of the etf at t - 1 plus
        # the sum of the changes in the instruments from t-1 to t accounting
        # for exchange rates, transaction costs, and dividends.
        etf[t] = Decimal(etf[t - 1]) + portfolio_sum

    return pd.Series(etf, index=index)


def portfolio_to_returns(portfolio: pd.DataFrame, timezone: str) -> pd.Series:
    prices = etf(portfolio, timezone)
    return prices_to_daily_returns(prices)


def prices_to_daily_returns(prices: pd.Series) -> pd.Series:
    """
    Calculates daily returns for a Series of prices.
    Note: drops the first value of the given Series.
    """
    return (prices / prices.shift(1) - 1)[1:]


def positions_to_dataframe(positions: Iterable[Position]) -> pd.DataFrame:
    """
    Returns a dataframe of positions with an additional `value` and `allocation` columns
    calculated from the averagePrice and quantity.
    """
    is_stock = lambda position: type(position.instrument) == Stock
    stocks = []
    for p in filter(is_stock, positions):
        stock = asdict(p)
        avgPrice = p.averagePrice
        stock["averagePrice"] = avgPrice.quantity
        stock["currency"] = avgPrice.currency
        stocks.append(stock)

    frame = pd.DataFrame.from_dict(stocks)
    frame["value"] = frame["averagePrice"] * frame["quantity"]
    frame["allocation"] = frame["value"] / frame["value"].sum()
    return frame


def positions_to_returns(
    provider: MarketDataProvider, positions: Iterable[Position], timezone: str
) -> pd.Series:
    frame = positions_to_dataframe(positions)
    positions, frame, history = positions_to_history(provider, positions, frame)
    return positions_and_history_to_returns(frame, history, timezone)


def positions_and_history_to_returns(
    frame: pd.DataFrame, historical_data: List[pd.DataFrame], timezone: str
) -> pd.Series:
    """
    Returns a Series of returns calculated by allocating $1 to the given historical data assets by the allocations specified in a positions frame.
    The timezones of the returns series are localized to the given timezone.
    """
    portfolio = positions_to_portfolio(frame, historical_data, timezone)
    return portfolio_to_returns(portfolio, timezone)


def positions_to_portfolio(
    frame: pd.DataFrame, historical_data: List[pd.DataFrame], timezone: str
) -> pd.DataFrame:
    """
    Returns a DataFrame of position histories with weights and allocation columns.
    """
    weights = {}
    components = {}
    for i, row in frame.reset_index().iterrows():
        weights[row["instrument"]["symbol"]] = row["allocation"]
        components[row["instrument"]["symbol"]] = historical_data[i]

    return stocks_to_portfolio(components, weights)


def positions_to_history(
    provider: MarketDataProvider, positions: Iterable[Position], frame: pd.DataFrame
) -> Tuple[List[Position], pd.DataFrame, List[pd.DataFrame]]:
    """
    Returns 1 year of daily historical data for a dataframe of positions.
    """
    is_stock = lambda position: type(position.instrument) == Stock
    bars = []
    new_positions = []
    for position in filter(is_stock, positions):
        try:
            # FIXME: This is specific to IBDataProvider right now
            barList = provider.fetchHistoricalData(position.instrument)  # type: ignore
        except ValueError:
            print("something bad happened")
            continue
        except:
            print("caught an exception")
            continue

        if barList is not None:
            new_positions.append(position)
        bars.append(barList)

    indices = [i for i, x in enumerate(bars) if x is not None]
    return (
        new_positions,
        frame.loc[indices],
        list(filter(lambda x: x is not None, bars)),
    )


def holdings(
    val: pd.DataFrame, holds: np.ndarray, i: pd.DataFrame, t: int, aum_t: Decimal
) -> Decimal:
    open_price = Decimal(val[i].loc["open"][t])

    # If the open price is NaN, this instrument's open wasn't recorded at time t.
    # So let's use the previous day's calculation.
    if not open_price.is_finite():
        prev_day: Decimal = holds[t - 1][val.columns.get_loc(i)]
        return prev_day
    else:
        # TODO: make the exchange rate flexible. This should generally be the dollar value of 1 point of instrument i.
        exchange_rate = Decimal(1)
        # The purpose of the (weights_i_t * sum_of_weights ^ -1) in the holdings calculation is to de-lever the allocations.
        sum_of_weights: Decimal = val.loc["weight"].iloc[t].abs().sum()

        # If today was t + 1, one way to calculate today's holdings would be by using the opening price, since for a future
        # we may not know the closing price of a new contract at roll time.
        # If open prices are unavailable, then the last close price at t will work too.
        next_open = Decimal(val[i].loc["close"][t])
        if not next_open.is_finite():
            last_close_price: Decimal = holds[t - 1][val.columns.get_loc(i)]
            return last_close_price

        weighted_holding: Decimal = Decimal(val[i].loc["weight"][t]) * aum_t / Decimal(
            next_open
        ) * exchange_rate * Decimal(sum_of_weights)
        return weighted_holding


def delta(val: pd.DataFrame, i: str, t: int) -> Decimal:
    """
    This function measures the change in market value from t - 1 to t.
    """
    before: Decimal = Decimal(val[i].loc["close"][t - 1])
    after: Decimal = Decimal(val[i].loc["close"][t])
    if math.isnan(before) or math.isnan(after):
        return Decimal(0)

    return after - before


def stocks_to_portfolio(
    components: Dict[str, pd.DataFrame], weights: Dict[str, float]
) -> pd.DataFrame:
    instruments = []
    for key, val in components.items():
        arrays = [
            ["date", "open", "high", "low", "close", "volume", "barCount", "average"],
            pd.DatetimeIndex(val["date"]),
        ]
        index = pd.MultiIndex.from_product(arrays, names=["field", "date"])
        df = pd.DataFrame(val.unstack().values, index=index)
        df.columns = [key]
        arrays = [
            [
                "date",
                "open",
                "high",
                "low",
                "close",
                "volume",
                "barCount",
                "average",
                "weight",
            ],
            pd.DatetimeIndex(val["date"]),
        ]
        df = df.reindex(pd.MultiIndex.from_product(arrays, names=["field", "date"]))
        df[key].loc["weight"] = np.repeat(
            weights[key], df[key].loc["weight"].index.shape[0]
        )
        instruments.append(df)
    return pd.concat(instruments, join="inner", sort=False, axis=1).sort_index()



from bankroll.broker import AccountAggregator
from bankroll.broker.configuration import Configuration
from bankroll.marketdata import MarketDataProvider, MarketConnectedAccountData
from typing import Iterable

import pkg_resources


def loadConfig(
    searchPaths: Iterable[str] = Configuration.defaultSearchPaths
) -> Configuration:
    defaultConfigName = "bankroll.default.ini"
    defaultConfig = pkg_resources.resource_string(
        "bankroll.interface", defaultConfigName
    ).decode()

    return Configuration(
        searchPaths=searchPaths,
        defaultConfig=defaultConfig,
        defaultConfigName=defaultConfigName,
    )


def marketDataProvider(accounts: AccountAggregator) -> MarketDataProvider:
    return next(
        (
            account.marketDataProvider
            for account in accounts.accounts
            if isinstance(account, MarketConnectedAccountData)
        )
    )



from .analysis import (
    TimelineEntry,
    activityAffectsSymbol,
    convertCashToCurrency,
    currencyConversionRates,
    deduplicatePositions,
    liveValuesForPositions,
    normalizeInstrument,
    normalizeSymbol,
    realizedBasisForSymbol,
    timelineForSymbol,
)
from .portfolio import (
    delta,
    etf,
    holdings,
    portfolio_to_returns,
    positions_and_history_to_returns,
    positions_to_dataframe,
    positions_to_history,
    positions_to_portfolio,
    positions_to_returns,
    prices_to_daily_returns,
    stocks_to_portfolio,
)

__all__ = [
    "normalizeSymbol",
    "normalizeInstrument",
    "activityAffectsSymbol",
    "realizedBasisForSymbol",
    "TimelineEntry",
    "timelineForSymbol",
    "liveValuesForPositions",
    "deduplicatePositions",
    "currencyConversionRates",
    "convertCashToCurrency",
    "etf",
    "portfolio_to_returns",
    "prices_to_daily_returns",
    "positions_to_dataframe",
    "positions_to_returns",
    "positions_and_history_to_returns",
    "positions_to_portfolio",
    "positions_to_history",
    "holdings",
    "delta",
    "stocks_to_portfolio",
]



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="bankroll",
    version="0.4.1",
    author="Justin Spahr-Summers",
    author_email="justin@jspahrsummers.com",
    description="Ingest portfolio and other data from multiple brokerages, and analyze it",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/bankroll-py/bankroll",
    packages=["bankroll.analysis", "bankroll.interface"],
    package_data={
        "bankroll.analysis": ["py.typed"],
        "bankroll.interface": ["bankroll.default.ini", "py.typed"],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Framework :: Jupyter",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3",
        "Topic :: Office/Business :: Financial :: Investment",
        "Typing :: Typed",
    ],
    install_requires=[
        "bankroll_marketdata ~= 0.4.0",
        "bankroll_model ~= 0.4.0",
        "bankroll_broker ~= 0.4.1",
        "numpy ~= 1.17.0",
        "progress ~= 1.5",
        "pyfolio >= 0.9.2",
    ],
    extras_require={
        "ibkr": ["bankroll_broker_ibkr ~= 0.4.0"],
        "schwab": ["bankroll_broker_schwab ~= 0.4.0"],
        "fidelity": ["bankroll_broker_fidelity ~= 0.4.0"],
        "vanguard": ["bankroll_broker_vanguard ~= 0.4.0"],
    },
    keywords="trading investing finance portfolio",
    entry_points={"console_scripts": ["bankroll = bankroll.interface.__main__:main"]},
)



import operator
import re
from dataclasses import dataclass, replace
from datetime import datetime
from decimal import Decimal
from functools import reduce
from itertools import groupby
from typing import Dict, Iterable, Optional, Sequence, Tuple

from progress.bar import Bar  # type: ignore

from bankroll.marketdata import MarketDataProvider
from bankroll.model import (
    Activity,
    Cash,
    CashPayment,
    Currency,
    Forex,
    Instrument,
    Option,
    Position,
    Quote,
    Stock,
    Trade,
)


# Different brokers represent "identical" symbols differently, and they can all
# be valid. This function normalizes them so they can be compared across time
# and space.
def normalizeSymbol(symbol: str) -> str:
    # These issues mostly show up with separators for multi-class shares (like BRK A and B)
    return re.sub(r"[\.\s/]", "", symbol)


# Performs a similar operation to normalizeSymbol(), but lifted over
# Instruments (where it makes sense).
def normalizeInstrument(instrument: Instrument) -> Instrument:
    if isinstance(instrument, Stock):
        return Stock(
            symbol=normalizeSymbol(instrument.symbol),
            currency=instrument.currency,
            exchange=instrument.exchange,
        )
    elif isinstance(instrument, Option):
        # Handles the FutureOption subclass correctly as well.
        return replace(instrument, underlying=normalizeSymbol(instrument.underlying))
    else:
        return instrument


# Attempts to determine whether the given Activity concerns the provided
# symbol, after normalizing to a form that should permit comparison across
# brokers.
def activityAffectsSymbol(activity: Activity, symbol: str) -> bool:
    normalized = normalizeSymbol(symbol)

    if isinstance(activity, CashPayment):
        return (
            activity.instrument is not None
            and normalizeSymbol(activity.instrument.symbol) == normalized
        )
    elif isinstance(activity, Trade):
        return (
            isinstance(activity.instrument, Option)
            and normalizeSymbol(activity.instrument.underlying) == normalized
        ) or normalizeSymbol(activity.instrument.symbol) == normalized
    else:
        return False


# Calculates the "realized" basis for a particular symbol, given a trade
# history. This refers to the actual amounts paid in and out, including
# dividend payments, as well as money gained or lost on derivatives related to
# that symbol (e.g., short puts, covered calls).
#
# The principle here is that we want to treat dividends and options premium as
# "gains," where cost basis gets reduced over time as proceeds are paid out.
# This is not how the tax accounting works, of course, but it provides a
# different view into the return/profitability of an investment.
def realizedBasisForSymbol(symbol: str, activity: Iterable[Activity]) -> Optional[Cash]:
    def f(basis: Optional[Cash], activity: Activity) -> Optional[Cash]:
        if isinstance(activity, CashPayment):
            return basis - activity.proceeds if basis else -activity.proceeds
        elif isinstance(activity, Trade):
            return basis - activity.proceeds if basis else -activity.proceeds
        else:
            raise ValueError(f"Unexpected type of activity: {activity}")

    return reduce(f, (t for t in activity if activityAffectsSymbol(t, symbol)), None)


@dataclass(frozen=True)
class TimelineEntry:
    date: datetime
    positions: Dict[Instrument, Decimal]
    realizedProfit: Cash

    def __str__(self) -> str:
        strPositions = (
            f"\n\t{instrument:21} {self.positions[instrument].normalize():>14,f}"
            for instrument in sorted(self.positions.keys())
        )

        return reduce(
            operator.add,
            strPositions,
            f"As of {self.date.date()}: {self.realizedProfit}",
        )


# Traces position sizing and profit/loss of a particular symbol over a period
# of activity. Yields a TimelineEntry corresponding to each action that
# occurred to the given symbol, starting from the oldest and ending with the
# most recent.
def timelineForSymbol(symbol: str, a: Iterable[Activity]) -> Iterable[TimelineEntry]:
    realizedProfit: Optional[Cash] = None
    positions: Dict[Instrument, Decimal] = {}

    for t in sorted(
        (t for t in a if activityAffectsSymbol(t, symbol)), key=lambda t: t.date
    ):
        if isinstance(t, CashPayment) or isinstance(t, Trade):
            proceeds = t.proceeds
        else:
            raise ValueError(f"Unexpected type of activity: {t}")

        if realizedProfit:
            realizedProfit += proceeds
        else:
            realizedProfit = proceeds

        if isinstance(t, Trade):
            instrument = normalizeInstrument(t.instrument)

            newPosition = positions.get(instrument, Decimal(0)) + t.quantity
            if newPosition == Decimal(0):
                del positions[instrument]
            else:
                positions[instrument] = newPosition

        yield TimelineEntry(
            date=t.date, positions=positions.copy(), realizedProfit=realizedProfit
        )


def liveValuesForPositions(
    positions: Iterable[Position],
    dataProvider: MarketDataProvider,
    progressBar: Optional[Bar] = None,
) -> Dict[Position, Cash]:
    def priceFromQuote(q: Quote, p: Position) -> Optional[Cash]:
        # For a long position, the value should be what the market is willing to pay right now.
        # For a short position, the value should be what the market is asking to be paid right now.
        if p.quantity < 0:
            return q.ask or q.last or q.bid or q.close
        else:
            return q.bid or q.last or q.ask or q.close

    result = {}

    positionsByInstrument: Dict[Instrument, Position] = {}
    for p in positions:
        if p.instrument in positionsByInstrument:
            raise ValueError(
                f"Expected unique instruments (i.e., deduplicated positions), but saw {p.instrument} multiple times"
            )

        positionsByInstrument[p.instrument] = p

    quotes = dataProvider.fetchQuotes(positionsByInstrument.keys())
    it = progressBar.iter(quotes) if progressBar else quotes

    for (instrument, quote) in it:
        position = positionsByInstrument[instrument]
        price = priceFromQuote(quote, position)
        if not price:
            continue

        result[position] = price * position.quantity * instrument.multiplier

    return result


def deduplicatePositions(positions: Iterable[Position]) -> Iterable[Position]:
    return (
        reduce(operator.add, ps)
        for i, ps in groupby(
            sorted(positions, key=lambda p: p.instrument), key=lambda p: p.instrument
        )
    )


# Looks up how much each of the `otherCurrencies` cost in terms of
# `quoteCurrency`.
#
# Returns each of the other currencies (possibly out-of-order), along with a
# cash price denominated in the `quoteCurrency`. If a quote is not available
# for some reason, it is omitted from the results.
def currencyConversionRates(
    quoteCurrency: Currency,
    otherCurrencies: Iterable[Currency],
    dataProvider: MarketDataProvider,
) -> Iterable[Tuple[Currency, Cash]]:
    instruments = (
        Forex(
            baseCurrency=min(currency, quoteCurrency),
            quoteCurrency=max(currency, quoteCurrency),
        )
        for currency in otherCurrencies
    )

    return (
        (instrument.baseCurrency, quote.market)
        if instrument.quoteCurrency == quoteCurrency
        else (
            instrument.quoteCurrency,
            Cash(
                currency=instrument.baseCurrency,
                # FIXME: This unfortunately does not retain much precision when
                # dividing by JPY in particular (where the integral portion can
                # be quite large).
                # See https://github.com/bankroll-py/bankroll/issues/37.
                quantity=Decimal(1) / quote.market.quantity,
            ),
        )
        for instrument, quote in dataProvider.fetchQuotes(instruments)
        if quote.market and isinstance(instrument, Forex)
    )


# Converts the given cash values into `quoteCurrency` using forex market quotes.
def convertCashToCurrency(
    quoteCurrency: Currency, cash: Sequence[Cash], dataProvider: MarketDataProvider
) -> Cash:
    currencyRates = dict(
        currencyConversionRates(
            quoteCurrency=quoteCurrency,
            otherCurrencies=(c.currency for c in cash if c.currency != quoteCurrency),
            dataProvider=dataProvider,
        )
    )
    currencyRates[quoteCurrency] = Cash(currency=quoteCurrency, quantity=Decimal(1))

    for c in cash:
        if not c.currency in currencyRates:
            raise RuntimeError(
                f"Unable to fetch currency rate for {c.currency} to convert {c}"
            )

    return reduce(
        operator.add,
        (
            Cash(
                currency=quoteCurrency,
                quantity=c.quantity * currencyRates[c.currency].quantity,
            )
            for c in cash
        ),
        Cash(currency=quoteCurrency, quantity=Decimal(0)),
    )
